/*
 *  Copyright 2003-2014  Mikhail Kryshen <mikhail@kryshen.net>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.games1729.shariks;

import java.awt.*;
import java.awt.geom.*;
import java.awt.event.*;
import java.util.*;

class Field extends KeyAdapter {
    public static final int GAME_NOT_STARTED = -1;
    public static final int GAME_PAUSED = 0;
    public static final int GAME_MENU = 1;
    public static final int GAME_FINISHED = 2; 
    public static final int GAME_RUNNING = 3;

    public static final long WINNER_CHECK_TIME = 1000000000; // 1s

    final float width;
    final float height;
    
    final Game game;
    private Player[] players = new Player[0];

    private java.util.List<Sharik> shariks = new ArrayList<Sharik>();
    private java.util.List<Sharik> shuffledShariks;

    final Player NOBODY = new Player(this, Color.WHITE, "Nobody");

    private int state = GAME_NOT_STARTED;
    private boolean resumed = false; // recover from pause.   

    private long timeSinceWinnerCheck = Long.MIN_VALUE;
    private Player candidate = null;

    Field(Game game) {
	width = 1000f; height = 700f;
	this.game = game;
    }

    public void keyPressed(KeyEvent e) {
	switch (e.getKeyCode()) {
	case 27: // ESC
	    if (state == GAME_RUNNING ||
		state == GAME_PAUSED) {
		
		changeGameState(GAME_MENU);
		game.displayMenu();
	    }
	    break;
	case 19: // Pause
	    if (state == GAME_RUNNING) {				
		changeGameState(GAME_PAUSED);
		game.renderer.setMessage("*** Pause ***");
	    } else if (state == GAME_PAUSED) {
		game.renderer.setMessage(null);
		changeGameState(GAME_RUNNING);
	    }
	    break;	    
// 	case 113: // F2
// 	    // Save game (TESTING)
// 	    synchronized (this) {
// 		try {
// 		    java.io.ObjectOutputStream out = 
// 			new java.io.ObjectOutputStream
// 			(new java.io.FileOutputStream("shariks.ser"));
// 		    out.writeObject(shariks);
// 		    out.writeObject(players);
				
// 		    component.setMessage("Game saved.", 1000);
// 		} catch (Exception exc) {
// 		    exc.printStackTrace();
// 		}
// 	    }
// 	    break;
// 	case 114: // F3
// 	    // Load game (TESTING)
// 	    synchronized (this) {
// 		try {
// 		    java.io.ObjectInputStream in = 
// 			new java.io.ObjectInputStream
// 			(new java.io.FileInputStream("shariks.ser"));
// 		    shariks = (java.util.List)in.readObject();
// 		    players = (Player[])in.readObject();

// 		    for (Iterator i = shariks.iterator(); i.hasNext(); )
// 			((Sharik)i.next()).setField(this);
// 		    shuffledShariks = new ArrayList(shariks);
				
// 		    component.setMessage("Game loaded.", 1000);
// 		} catch (Exception exc) {
// 		    exc.printStackTrace();
// 		}
// 	    }
// 	    break;
	    
	}
    }

    synchronized void render(Graphics2D g) {
	float scale = getScale();

	for (Sharik s1 : shariks) {
	    if (s1 instanceof TailSharik) {
		Sharik s2 = ((TailSharik)s1).link;
		if (s2 != null) {
		    int alpha =
			255 - (int)(Math.sqrt
				    ((s1.x - s2.x) * (s1.x - s2.x) + 
				     (s1.y - s2.y) * (s1.y - s2.y)) *
				    255f / 100f);
		    if (alpha > 0) {
			g.setColor(new Color(255, 255, 255, alpha));
			g.draw(new Line2D.Float(s1.x * scale, s1.y * scale,
						s2.x * scale, s2.y * scale));
		    }
		} else {
		    g.setColor(new Color(255, 255, 255, 127));
		    g.draw(new Line2D.Float((s1.x - s1.r - 1f) * scale, 
					    (s1.y - s1.r - 1f) * scale,
					    s1.x * scale, s1.y * scale));
		}
	    }
	}

	for (Sharik s : shariks) {
	    int x = (int)(s.x * scale);
	    int y = (int)(s.y * scale);

	    g.translate(x, y);
	    s.look.render(s, g, 0);
	    g.translate(-x, -y);
	}
    }

    float getScale() {
	return Math.min((float)game.getWidth() / width,
			(float)game.getHeight() / height);
    }

    /**
     * Start new game.
     */
    synchronized void start(Player[] players) {
	this.players = players;
	
	shariks.clear();
	
	if (players.length == 1) {
	    for (int i = 0; i <= 2; i++) {
		shariks.add(new CrazySharik(this));
	    }

	    //for (int i = 0; i <= 2; i++)
	    //for (int i = 0; i <= 10 * players.length - 1; i++)
	    for (int i = 0; i <= 19; i++)
		shariks.add(new TailSharik(this));
	    	    
	} else {
	    for (int i = 0; i <= 0; i++) {
		shariks.add(new CrazySharik(this));
	    }

	    for (int i = 0; i <= 10; i++)
		shariks.add(new TailSharik(this));
	}

	for (int i = 0; i <= players.length - 1; i++)
	    shariks.add(players[i]);
	
	shuffledShariks = new ArrayList<Sharik>(shariks);

	// set initial locations
	for (int i = 0; i < shariks.size(); i++) {
	    boolean free;      
	    Sharik s1 = shariks.get(i);

	    do {
		s1.x = (width - s1.r * 2) * (float)Math.random() + s1.r;
		s1.y = (height - s1.r * 2) * (float)Math.random() + s1.r;
		
		free = true;
		for (int j = 0; j < i; j++) {
		    Sharik s2 = shariks.get(j);

		    if ((s2.x - s1.x) * (s2.x - s1.x) + 
			(s2.y - s1.y) * (s2.y - s1.y) <=
			(s2.r + s1.r) * (s2.r + s1.r)) {
			
			free = false;
			break;
		    }
		}
	    } while (!free);
	}

	changeGameState(GAME_RUNNING);
    }

    synchronized void changeGameState(int state) {	
	this.state = state;
	switch (state) {
	case GAME_RUNNING:
	    resume();
	    game.setLazyRenderMode(false);
	    break;
	case GAME_PAUSED:
	case GAME_MENU:
	case GAME_FINISHED:
	    stop();
	    game.setLazyRenderMode(true);
	    break;
	}
    }

    public int getGameState() {
	return state;
    }

    private void stop() {	
	for (int i = 0; i <= players.length - 1; i++) {
	    players[i].stop();
	}
    }

    /**
     * Resume stoped game.
     */
    private void resume() {
	for (int i = 0; i <= players.length - 1; i++) {
	    players[i].start();
	}
	resumed = true;
    }

    private static final int BORDER_VERTICAL = -1;
    private static final int BORDER_HORIZONTAL = -2;       

    void update(long lag) {
	if (state != GAME_RUNNING)
	    return;

	if (resumed) {
	    // Skip one update after resuming pause game.
	    resumed = false;
	    return;
	}

	updatePositions(lag);

	timeSinceWinnerCheck += lag;

	if (timeSinceWinnerCheck > Long.MIN_VALUE + WINNER_CHECK_TIME) {
	    timeSinceWinnerCheck = Long.MIN_VALUE;

	    if (candidate == null) {
		candidate = findWinner();
	    } else {
		Player oldCandidate = candidate;
		candidate = findWinner();

		if (oldCandidate == candidate) {
		    changeGameState(GAME_FINISHED);		    
		    game.renderer.setMessage(candidate.getName() + " wins!");

		    // Display menu after 3 seconds.
		    (new Thread() {
			    public void run() {
				try {
				    sleep(3000);
				} catch (InterruptedException e) {
				    return;
				}
				game.displayMenu();
			    }
			}).start();
		}
	    }
	}
    }
    
    private Player findWinner() {
	Player candidate = null;
	TailSharik ts = null;
	int n = 0;
	boolean b = false;

	for (Sharik s : shariks) {
	    if (s instanceof TailSharik) {
		ts = (TailSharik)s;

		Player player = null;
		for (int j = 0; j <= players.length - 1; j++) {
		    if (ts.look == players[j].look)
			player = players[j];
		}

		if (ts.link == player) {
		    b = true;
		    ts.linked = true;
		} else {
		    ts.linked = false;
		    n++;
		}

		if (player == null)
		    return null;
		if (candidate == null)
		    candidate = player;
		else if (candidate != player)
		    return null;
	    }
	}
	
	if (!b) {
	    ts.linked = true;
	    n--;
	}

	for (int i = 0; i <= shariks.size() - 1; i++) {
	    for (Sharik s : shariks) {
		if (s instanceof TailSharik) {
		    ts = (TailSharik)s;
		    if (ts.link instanceof TailSharik &&
			(((TailSharik)ts.link).linked || ts.linked) &&
			!(((TailSharik)ts.link).linked && ts.linked)) {
			
			n--;
			ts.linked = ((TailSharik)ts.link).linked = true;
		    }
		}
	    }
	}

	return (n == 0) ? candidate : null;
    }

    private void updatePositions(long lag) {
	float t = (float)lag * 1E-6f;

	int n = shariks.size();

	Collections.shuffle(shuffledShariks, Game.random);

	int cid = 1;

	// Calculate accelerations.
	for (int i = 0; i < n; i++) {
	    Sharik s1 = shuffledShariks.get(i);

	    Vector2D f = s1.getDrivingForce();
	    f.add(-s1.vx * s1.drag, -s1.vy * s1.drag);

	    for (int j = 0; j <= n - 1; j++) if (i != j) {
		Sharik s2 = shuffledShariks.get(j);

		f.add(s2.getAttractingForce(s1));
		f.sub(s1.getAttractingForce(s2));
	    }

	    s1.ax = f.x / s1.m;
	    s1.ay = f.y / s1.m;

	    s1.ct = -1f;
	    s1.cid = 0;
	}

	// Resolve collisons, calculate velocity and location for each Sharik.
	float dt = 0;
	while (dt < t) {
	    float ct = t - dt; // Time to the next collision.
	    Sharik cs1 = null;
	    Sharik cs2 = null;

	    for (int i = 0; i < n; i++) {
		Sharik s1 = shuffledShariks.get(i);

		// Sharik-sharik collisions.
		for (int j = i + 1; j < n; j++) {
		    Sharik s2 = shuffledShariks.get(j);

		    float a = (s1.vx - s2.vx) * (s1.vx - s2.vx) + 
			      (s1.vy - s2.vy) * (s1.vy - s2.vy);

		    if (a != 0) {
			float b = (s1.vx - s2.vx) * (s1.x - s2.x) + 
			          (s1.vy - s2.vy) * (s1.y - s2.y);
			float c = (s1.x - s2.x) * (s1.x - s2.x) + 
			          (s1.y - s2.y) * (s1.y - s2.y) - 
			          (s1.r + s2.r) * (s1.r + s2.r);
			
			// Solve a*t^2 + 2*b*t + c = 0.
			float d = b * b - a * c;
			
			if (d >= 0) {
			    float t1 = (-b - (float)Math.sqrt(d)) / a;
			    float t2 = (-b + (float)Math.sqrt(d)) / a;
			    
			    if (t2 >= 0 && t2 < t1) t1 = t2;
			    
			    if (t1 < ct && (t1 > 0 || 
					    (t1 == 0 && (s1.ct < dt + t1 || 
							 s2.ct < dt + t1 ||
							 s1.cid != s2.cid)))) {
				ct = t1;			    
				cs1 = s1;
				cs2 = s2;			    
			    }
			}
		    }
		}

		// Sharik-border collisions.
		float t1;

		if (s1.vx < 0) {
		    t1 = (s1.r - s1.x) / s1.vx;
		    if (t1 < ct && (t1 > 0 || (t1 == 0 && (s1.ct < dt + t1 || 
							   s1.cid != BORDER_VERTICAL)))) {
			ct = t1;
			cs1 = s1;
			cs2 = null;
		    }
		} else if (s1.vx > 0) {
		    t1 = (width - s1.r - s1.x) / s1.vx;
		    if (t1 < ct && (t1 > 0 || (t1 == 0 && (s1.ct < dt + t1 || 
							   s1.cid != BORDER_VERTICAL)))) {
			ct = t1;
			cs1 = s1;
			cs2 = null;
		    }
		}
		
		if (s1.vy < 0) {
		    t1 = (s1.r - s1.y) / s1.vy;
		    if (t1 < ct && (t1 > 0 || (t1 == 0 && (s1.ct < dt + t1 || 
							   s1.cid != BORDER_HORIZONTAL)))) {
			ct = t1;
			cs1 = null;
			cs2 = s1;
		    }		    
		} else if (s1.vy > 0) {
		    t1 = (height - s1.r - s1.y) / s1.vy;
		    if (t1 < ct && (t1 > 0 || (t1 == 0 && (s1.ct < dt + t1 || 
							   s1.cid != BORDER_HORIZONTAL)))) {
			ct = t1;
			cs1 = null;
			cs2 = s1;
		    }
		}
	    }

	    dt += ct;

	    // Calculate Shariks states at the time of the next collision.
	    for (int i = 0; i < n; i++) {
		Sharik s = shuffledShariks.get(i);

		s.x += s.vx * ct;
		s.y += s.vy * ct;
		s.vx += s.ax * ct;
		s.vy += s.ay * ct;		    
	    }
	    
	    // Calculate velocity after collision.
	    if (cs1 != null && cs2 != null) {
		float rx = cs1.x - cs2.x;
		float ry = cs1.y - cs2.y;

		float lr = (float)Math.sqrt(rx * rx + ry * ry);
		
		float v1 = (rx * cs1.vx + ry * cs1.vy) / lr;
		float v2 = (rx * cs2.vx + ry * cs2.vy) / lr;

		float vnx1 = cs1.vx - rx * v1 / lr;
		float vny1 = cs1.vy - ry * v1 / lr;

		float vnx2 = cs2.vx - rx * v2 / lr;
		float vny2 = cs2.vy - ry * v2 / lr;

		float vt = v1;
		v1 = ((cs1.m - cs2.m) * v1 + 2 * cs2.m * v2) / (cs1.m + cs2.m);
		v2 = ((cs2.m - cs1.m) * v2 + 2 * cs1.m * vt) / (cs2.m + cs1.m);
		
		cs1.vx = vnx1 + rx * v1 / lr;
		cs1.vy = vny1 + ry * v1 / lr;

		cs2.vx = vnx2 + rx * v2 / lr;
		cs2.vy = vny2 + ry * v2 / lr;

		cs1.ct = dt;
		cs2.ct = dt;
		cs1.cid = cs2.cid = cid++;
	    } else if (cs1 != null) {
		cs1.vx = -cs1.vx;

		cs1.ct = dt;
		cs1.cid = BORDER_VERTICAL;
	    } else if (cs2 != null) {
		cs2.vy = -cs2.vy;

		cs2.ct = dt;
		cs2.cid = BORDER_HORIZONTAL;
	    }
	}
    }
}
